/**********************************************************************************
21/November/2000 (c) SCEE. Written by Jason Page.

xmrpc.c
	Code for updating IOP XM Player module via RPC call
***********************************************************************************/

#include <eekernel.h>
#include <stdio.h>
#include <sdrcmd.h>
#include <libdev.h>
#include <libpkt.h>
#include <sif.h>
#include <sifrpc.h>
#include "../iop/xmcalls.h"
#include "xmrpc.h"

u_short senddata[128] __attribute__((aligned (64)));

int IOP_Cnt=0;		// Pointer into senddata for writing command/data packets


#define DATA_SIZE_NORMAL (128*2)

// RPCReturnBuffer holds the data that is returned from the IOP.
u_int RPCReturnBuffer[128] __attribute__((aligned (64)));

// These values are filled in from current playback of XM file..
int XM_CurrentPattern;
int XM_SongPos;
int XM_SongBPM;
int XM_SongSpeed;
int XM_SongStatus;
int XM_SongPatternPos;
int XM_SongLength;
int XM_SongChannels;
int XM_SongChannelVolumes[32];

sceSifClientData jgCd;

u_short IOPCommandData[128];

/**********************************************************************************
SOUND_InitIOP
	Initialise IOP sound routine
	Bind the reqiured IOP device (ie. link it into the system..)
***********************************************************************************/
int SOUND_InitIOP( void )
{
	int fd, i,cid;
	char *cp;

	sceSifInitRpc(0);

	while(1) {
		if (sceSifBindRpc( &jgCd, SOUND_DEV, 0) < 0) {
			printf("error: sceSifBindRpc \n");
			while(1);
		}
		i = 10000;		// looks a bit dodgy - eh?!
		while( i-- ){}
		if(jgCd.serve != 0) break;
	}
	senddata[0]=0;
	return 1;
}

/**********************************************************************************
SOUND_SIFTransfer
	Transfers data from EE to IOP

	Requires:	command		Command to execute on IOP
				data		address of data packet
				flag		0(halts EE)  or SIF_RPCM_NO_WAIT(doesn't stall EE)
***********************************************************************************/
int SOUND_SIFTransfer( int command,  int data, int flag )
{
int a;
int i, rsize;

	rsize = 128*4;
	sceSifCallRpc( &jgCd,  command, flag, (void *)data
	, DATA_SIZE_NORMAL, (void *)(&RPCReturnBuffer[0]),rsize,NULL,NULL);
	return RPCReturnBuffer[0];
}

/**********************************************************************************
FlushIOPCommand
	Transfers IOP command/data packet to the IOP via SIFDMA.

	Requires:	type	0=Wait for IOP to process commands
						SIF_RPCM_NOWAIT= No waiting.
	Returns:
				-1		RPC already running (ie. still processing data from last send)
				-2		No Data to be sent.
				oldcnt	Number of command/data packets sent
***********************************************************************************/
int FlushIOPCommand(int type)
{
int oldcnt,a;

	a=sceSifCheckStatRpc((sceSifRpcData*)&jgCd);		// Make sure there is not any Rpc's running already.
	if (a!=0)
		return(-1);

	if (senddata[0]==0)
		return(-2);
	SOUND_SIFTransfer( SOUND_SETPARAMS,(int)senddata,type);			// set rev speed from left analog stick
	oldcnt=IOP_Cnt;
	senddata[0]=0;
	IOP_Cnt=0;
	return (oldcnt);
}

/*****************************************************************************
XM_SetSPUBaseAddress
	Sets address where samples will be transfered to during the XM_TransferSampleData
	call.
*****************************************************************************/
void XM_SetSPUBaseAddress(unsigned int addr)
{
unsigned int ad1,ad2;

	ad1=addr>>16;
	ad2=addr&0xffff;

	IOPCommandData[0]=ad1;
	IOPCommandData[1]=ad2;
	SetIOPCommand(XMI_SetTransAddress,2);
}

/*****************************************************************************
XM_CreateIOPTransferBuffer
	Creates 64K IOP->SPU Transferbuffer RAM
*****************************************************************************/
void XM_CreateIOPTransferBuffer(void)
{
	SetIOPCommand(XMI_CreateTransBuff,0);
}

/*****************************************************************************
XM_FreeIOPTransferBuffer
	Free 64K IOP->SPU Transferbuffer RAM
*****************************************************************************/
void XM_FreeIOPTransferBuffer(void)
{
	SetIOPCommand(XMI_FreeTransBuff,0);
}

/*****************************************************************************
XM_AlloceXMBuffer
	Allocate IOP RAM for XMBuffer. Relevent parts of the XM file are copied
	into this when calling XM_InitXMData, for fast/easy access later on.
*****************************************************************************/
void XM_AllocXMBuffer(int id)
{
	IOPCommandData[0]=id;
	SetIOPCommand(XMI_AllocXMBuffer,1);
}

/*****************************************************************************
XM_AlloceXMSong
	Allocate IOP RAM for using as a song buffer. Up to 32 songs can be played
	at once. This buffer holds info for current playback / channel info.etc.
*****************************************************************************/
void XM_AllocXMSong(int SongNumber)
{
	IOPCommandData[0]=SongNumber;
	SetIOPCommand(XMI_AllocXMSong,1);
}

/*****************************************************************************
XM_AllocateXMPatternRAM
	Allocate IOP RAM for transfering XM file to
*****************************************************************************/
void XM_AllocXMPatternRAM(int size,int slot)
{
unsigned int sizea,sizeb;
	sizea=size>>16;
	sizeb=size&0xffff;
	IOPCommandData[0]=sizea;
	IOPCommandData[1]=sizeb;
	IOPCommandData[2]=slot;
	SetIOPCommand(XMI_AllocXMPatternRAM,3);
}

/*****************************************************************************
XM_InitXMData
	Initialises XM data, copying relevent parts from the XM file into the
	XM Buffer
*****************************************************************************/
void XM_InitXMData(int xm_id, int pantype)
{
	IOPCommandData[0]=xm_id;
	IOPCommandData[1]=pantype;
	SetIOPCommand(XMI_InitXMData,2);
}

/*****************************************************************************
XM_XMInit
	Once off call (do at the start of the program...)
	Initialises some buffers and sets stereo mode as default
*****************************************************************************/
void XM_XMInit(void)
{
	SetIOPCommand(XMI_XMInit,0);
}
/*****************************************************************************
XM_SetMasterVolume
	Sets master volume position of song. Can only be setup after XM_XMInit call..
*****************************************************************************/
void XM_SetMasterVolume(int songid,int vol)
{
	IOPCommandData[0]=songid;
	IOPCommandData[1]=vol;
	SetIOPCommand(XMI_SetMasterVolume,2);
}

/*****************************************************************************
XM_PlayNextPos
	Causes the playback of a song to jump to the specified position when it
	has completed the current pattern (good for interactive music...)
*****************************************************************************/
void XM_PlayNextPos(int songid,int pos)
{
	IOPCommandData[0]=songid;
	IOPCommandData[1]=pos;			// -1 to cancel..
	SetIOPCommand(XMI_PlayNextPos,2);
}

/*****************************************************************************
XM_SetMasterPan
	Sets master pan position of song. Can only be setup after XM_XMInit call..
*****************************************************************************/
void XM_SetMasterPan(int songid,int pan)
{
	IOPCommandData[0]=songid;
	IOPCommandData[1]=pan;
	SetIOPCommand(XMI_SetMasterPan,2);
}

/*****************************************************************************
XM_CreateSampleOffsetData
	Sets address of xm instrument (in SPU RAM) in IOP RAM
	vid = SampleBank ID
	slot = instrument number
	address = SPU Address
*****************************************************************************/
void XM_CreateSampleOffsetData(int vid,int slot, unsigned int address)
{
unsigned int ad1,ad2;

	ad1=address>>16;
	ad2=address&0xffff;
	IOPCommandData[0]=vid;
	IOPCommandData[1]=slot;
	IOPCommandData[2]=ad1;
	IOPCommandData[3]=ad2;
//	printf("com %x\n",address);
	SetIOPCommand(XMI_SetSampleOffsets,4);
}

/*****************************************************************************
XM_TransferSampleData
	Transfer data from IOP 64K Buffer to SPU
	This will also update the current SPU transfer position on completion.
*****************************************************************************/
void XM_TransferSampleData(int size)
{
unsigned int sizea,sizeb;

	sizea=size>>16;
	sizeb=size&0xffff;
	IOPCommandData[0]=sizea;
	IOPCommandData[1]=sizeb;
	SetIOPCommand(XMI_TransferSample,2);
}

/*****************************************************************************
XM_XMPlay
	Play a song
	VAB_ID	= SampleBank ID
	XM_ID	= XM ID (!)
	SongID	= Song ID
	firstch = First SPU channel to use for playback (0-47)
	Loop	= XM_Loop (loops tune when reaches end) or XM_NoLoop (for a once off tune)
	offset	= Start position through XM Pattern data
*****************************************************************************/
void XM_XMPlay(int VAB_ID,int XM_ID,int SongID,int firstch, int Loop,int bitmask,int offset)
{
unsigned int ad1,ad2;

	ad1=bitmask>>16;
	ad2=bitmask&0xffff;

	IOPCommandData[0]=VAB_ID;
	IOPCommandData[1]=XM_ID;
	IOPCommandData[2]=SongID;
	IOPCommandData[3]=firstch;
	IOPCommandData[4]=Loop;
	IOPCommandData[5]=ad1;
	IOPCommandData[6]=ad2;
	IOPCommandData[7]=offset;
	SetIOPCommand(XMI_PlayXM,8);
}

/*****************************************************************************
XM_Pause
	Pause song (if playing)
*****************************************************************************/
void XM_Pause(int SongID)
{
	IOPCommandData[0]=SongID;
	SetIOPCommand(XMI_Pause,1);
}

/*****************************************************************************
XM_Resume
	Resume playing song (if paused)
*****************************************************************************/
void XM_Resume(int SongID)
{
	IOPCommandData[0]=SongID;
	SetIOPCommand(XMI_Resume,1);
}

/*****************************************************************************
XM_Exit
	Stop all songs
*****************************************************************************/
void XM_XMExit(void)
{
	SetIOPCommand(XMI_XMExit,0);
}

/*****************************************************************************
XM_XMQuit
	Stop song
*****************************************************************************/
void XM_XMQuit(int SongID)
{
	IOPCommandData[0]=SongID;
	SetIOPCommand(XMI_XMQuit,1);
}

/*****************************************************************************
XM_PauseAll
	Pause all playing songs
*****************************************************************************/
void XM_PauseAll(void)
{
	SetIOPCommand(XMI_Pause,0);
}

/*****************************************************************************
XM_ResumeAll
	Resume all paused songs
*****************************************************************************/
void XM_ResumeAll(void)
{
	SetIOPCommand(XMI_ResumeAll,0);
}

/*****************************************************************************
XM_FreeXMBuffer
	Free IOP RAM that holds a XM Buffer
*****************************************************************************/
void XM_FreeXMBuffer(int xm_id)
{
	IOPCommandData[0]=xm_id;
	SetIOPCommand(XMI_FreeXMBuffer,1);
}

/*****************************************************************************
XM_FreeXMSong
	Free IOP RAM that holds a song buffer
*****************************************************************************/
void XM_FreeXMSong(int song_id)
{
	IOPCommandData[0]=song_id;
	SetIOPCommand(XMI_FreeXMSong,1);
}

/*****************************************************************************
XM_FreeXMPatternRAM
	Free IOP RAM that holds the .XM File
*****************************************************************************/
void XM_FreeXMPatternRAM(int xm_id)
{
	IOPCommandData[0]=xm_id;
	SetIOPCommand(XMI_FreeXMPatternRAM,1);
}

/*****************************************************************************
XM_GetSongFeedback
	Fills the IOP Return buffer with relevent info for specified song
*****************************************************************************/
void XM_GetSongFeedback(int songid)
{
	IOPCommandData[0]=songid;
	SetIOPCommand(XMI_GetSongFeedback,1);
}

/*****************************************************************************
XM_SetSongPos;
	Set current playback position of song to specific place..
*****************************************************************************/
void XM_SetSongPos(int songid,int pos)
{
	IOPCommandData[0]=songid;
	IOPCommandData[1]=pos;
	SetIOPCommand(XMI_SetSongPos,2);
}

/*****************************************************************************
XM_SetChannelVolume
	Set single channel of song's volume (not SPU Channel..0 always first channel
	of a song..)
	Volume = 0 - 64 (64 = full volume)
	Note: These values are reset to full volume when a song is first played.
*****************************************************************************/
void XM_SetChannelVolume(int songid,int channel,int volume)
{
	IOPCommandData[0]=songid;
	IOPCommandData[1]=channel;
	IOPCommandData[2]=volume;
	SetIOPCommand(XMI_SetChannelVolume,3);
}

/*****************************************************************************
SOUND_InitSPU
	Init LibSD and set master volumes to full
*****************************************************************************/
void SOUND_InitSPU(void)
{
	SetIOPCommand(XMI_INIT_SPU,0);
}

/*****************************************************************************
XM_SetBPM
	Set BPM of song
*****************************************************************************/
void XM_SetBPM(int songid,int bpm)
{
	IOPCommandData[0]=songid;
	IOPCommandData[1]=bpm;
	SetIOPCommand(XMI_SetBPM,2);
}

/*****************************************************************************
XM_SetSpeed
	Set Speed of song
*****************************************************************************/
void XM_SetSpeed(int songid,int speed)
{
	IOPCommandData[0]=songid;
	IOPCommandData[1]=speed;
	SetIOPCommand(XMI_SetSpeed,2);
}

/*****************************************************************************
XM_SetStereo
	Set Playback to mono mode for ALL songs
*****************************************************************************/
void XM_SetStereo(void)
{
	SetIOPCommand(XMI_SetStereo,0);
}

/*****************************************************************************
XM_SetMono
	Set Playback to mono mode for ALL songs
*****************************************************************************/
void XM_SetMono(void)
{
	SetIOPCommand(XMI_SetMono,0);
}

/*****************************************************************************
XM_PlaySample
	Play a sample from sample bank.
	vid = Sample Bank ID
	slot = sample to play
	channel = 0-47
	voll / volr = volume (0-0x3fff)
	pitch = spu pitch
*****************************************************************************/
void XM_PlaySample(int vid,int slot,int channel,int voll,int volr,int pitch)
{
	IOPCommandData[0]=vid;
	IOPCommandData[1]=slot;
	IOPCommandData[2]=channel;
	IOPCommandData[3]=voll;
	IOPCommandData[4]=volr;
	IOPCommandData[5]=pitch;
	SetIOPCommand(XMI_PlaySample,6);
}

/*****************************************************************************
XM_StopSample
	Stop playing sample on channel (0-47)
*****************************************************************************/
void XM_StopSample(int ch)
{
	IOPCommandData[0]=ch;
	SetIOPCommand(XMI_StopSample,1);
}

/*****************************************************************************
XMEE_QuickInit
	Allocates XM IOP Buffer. (using xm_id)
	Allocates XM IOP Pattern Buffer. (using xm_id)
	Transfer XM File to IOP Pattern Buffer
	Allocate song buffer (using song_id)
	Initialise XM data

	Requires: panswitch 	(Either XM_UseXMPanning, or XM_UseS3MPanning)
	Returns: IOP address of XM File
*****************************************************************************/
u_int XMEE_QuickInit(int xm_id,int song_id,u_int source,u_int size,int panswitch)
{
u_int iopaddr;

// Alloc memory for storing xm info (max 8 xm files can be used at one time)
// These buffers are used for holding info about the xm files (original bpm,speed etc...)
// Although this data is already stored in the xm file, its not very easy to extract some of the info.
	XM_AllocXMBuffer(xm_id);

// Now allocate space on the IOP for transfering the whole of the XM file to..
	XM_AllocXMPatternRAM(size,xm_id);	// Allocate IOP RAM for XM file
	FlushIOPCommand(0);					// Wait for command to process..
	iopaddr=RPCReturnBuffer[2];			// Copy address to a safe place

//Transfer XM Pattern data to the IOP (the .XM file..)
	XMEE_TransferDataToIOP(source,iopaddr,size);

// Alloc memory for a song buffer in slot n (song buffer holds info for current playback)
// You can play up to 32 songs at once - played from a max of 8 XM files..
	XM_AllocXMSong(song_id);		// ID must range from 0-31

// Initialise XM pattern file (pull out all the relevent data like bpm/speed/number of channels...)
// panswitch is either XM_UseXMPanning or XM_UseS3MPanning (use if playing S3M files that were converted to XM)
	XM_InitXMData(xm_id, panswitch);
	FlushIOPCommand(0);
	return(iopaddr);			// Returns IOP Address of XM Pattern Data
}


/*****************************************************************************
XMEE_QuickShutdown
	Stops play song. Free IOP Song memory. Free XM IOP buffer memory
	Free XM IOP Pattern memory
*****************************************************************************/
void XMEE_QuickShutdown(int xm_id,int song_id)
{
	XM_XMQuit(song_id);			// Stop playing / processing songe
	XM_FreeXMSong(song_id);		// Free the song IOP RAM (allocatbbed by XM_AllocXMSong(SongID))
	XM_FreeXMBuffer(xm_id);		// Free IOP XM Buffer (holding important info extracted from xm file..)
	XM_FreeXMPatternRAM(xm_id);	// Free IOP XM Pattern RAM (where the actual file was transfered to.
	FlushIOPCommand(0);
}

/*****************************************************************************
XMEE_SetFeedback
	Copy values from IOP return buffer into more meaningful variable names.
*****************************************************************************/
void XMEE_SetFeedback(void)
{
int i,c;
	XM_CurrentPattern=RPCReturnBuffer[6];
	XM_SongPos=RPCReturnBuffer[7];
	XM_SongBPM=RPCReturnBuffer[8];
	XM_SongSpeed=RPCReturnBuffer[9];
	XM_SongStatus=RPCReturnBuffer[10];
	XM_SongPatternPos=RPCReturnBuffer[11];
	XM_SongLength=RPCReturnBuffer[12];
	XM_SongChannels=RPCReturnBuffer[13];

	c=0;
	for (i=0;i<16;i++)
	{
		XM_SongChannelVolumes[c]=(RPCReturnBuffer[14+i]>>8)&0x3f;
		XM_SongChannelVolumes[c+1]=RPCReturnBuffer[14+i]&0x3f;
		c+=2;
	}
}

/*****************************************************************************
XMEE_TransferSampleData
	Transfer sample data block from EE to IOP
*****************************************************************************/
u_char jb[64]__attribute__((aligned (64)));
int XMEE_TransferSampleData(int source,int dest,int size)
{
int size2;
int did;
sceSifDmaData JDMA;
int i;

	while (size>0)
	{
		if (size>=(64*1024))
		{
			size2=64*1024;		// Transfer a maximum of 64K each time
		}
		else
		{
			size2=size+63;
			size2&=(~63);		// Round last transfer to 64 bytes.
		}

		FlushCache(0);

		JDMA.data = source;
		JDMA.addr = dest;
		JDMA.size = size2;
		JDMA.mode = 0;
		if (!(did = sceSifSetDma(&JDMA, 1)))
		{
			printf("Sample data transfer failed..\n");
			return -1;
		}
	 	while(sceSifDmaStat(did) >= 0);

		XM_TransferSampleData(size2);	// Transfer buffer from IOP to SPU RAM
		FlushIOPCommand(0);		// Send command to IOP and wait for this to be done.
		size-=size2;
		source+=size2;
	}

	return(RPCReturnBuffer[0]);	// New SPU Base address (where, say, another bank of samples will be loaded to...)
}

/*****************************************************************************
XMEE_TransferDataToIOP
	Transfer data from EE to IOP (XM file)
*****************************************************************************/
int XMEE_TransferDataToIOP(unsigned int source,unsigned int dest, unsigned int size)
{
sceSifDmaData JDMA;
int did;

	FlushCache(0);
	JDMA.data = source;
	JDMA.addr = dest;
	JDMA.size = size;
	JDMA.mode = 0;
	if (!(did = sceSifSetDma(&JDMA, 1)))
	{
		printf("Transfer to IOP Failed\n");
		return -1;
	}
 	while(sceSifDmaStat(did) >= 0);
	return(0);
}

/*****************************************************************************
XMEE_CreateSampleOffsetData
	Send sizes of each sample in bank to IOP to build a list of sample start
	addresses
*****************************************************************************/
void XMEE_CreateSampleOffsetData(int vid, u_char *Vh)
{
u_short Vags,Progs;
u_short *dataptr;
u_int *datatype;
int i;
unsigned int addr;
int size;

	datatype=(u_int*)(Vh);
	i=*datatype;
	if (i!=0x56414270)		// "pBAV" (first long in a .VH file)
	{
		dataptr=(u_short*)(Vh+2);	// PS2 OUTPUT FROM XM2PSX TOOL..
		addr=RPCReturnBuffer[0];
		for (i=0;i<128;i++)
		{
			size=*dataptr;
			size=(size<<3);
			if (size!=0)
			{
				XM_CreateSampleOffsetData(vid,i,addr);
				addr+=size;
			}
			else
			{
				XM_CreateSampleOffsetData(vid,i,0);
			}
			FlushIOPCommand(0);		// Send commands to IOP and wait for them to be processed.
			dataptr++;
		}
	}
	else		// THIS IS CALLED IF THE FILE FORMAT USES THE OLD PS1 "VH/VB" DATA
	{
		Vags=(u_short)*(Vh+22);
		Progs=(u_short)*(Vh+18);
		addr=RPCReturnBuffer[0];
		dataptr=(short *)(Vh+(32+2048+((16*Progs)*32)+2));

		for (i=0;i<Vags;i++,dataptr++)
		{
			size=*dataptr;
			size=(size<<3);
			if (size!=0)
			{
				XM_CreateSampleOffsetData(vid,i,addr);
				addr+=size;
			}
			else
			{
				XM_CreateSampleOffsetData(vid,i,0);
			}
			FlushIOPCommand(0);		// Send commands to IOP and wait for them to be processed.
		}
	}
}

/**********************************************************************************
SetIOPCommand
	Builds an 8 short list of command/data, which is sent over to IOP once per frame.

	Requires:	Command		Command to execute (PlayStream/StopSound...etc..)
				Length		Length of data required for command
***********************************************************************************/
void SetIOPCommand(short command,short length)
{
int i;

	if ((IOP_Cnt+length+3)>=128)	// +1 for first short being the command count. +1 for command short.
	{
		printf ("*** Function:SetIOPCommand - IOP Buffer full ***\n");
		return;
	}
	senddata[IOP_Cnt+1]=command;
	senddata[IOP_Cnt+2]=length;

	for (i=0;i<length;i++)
	{
		senddata[IOP_Cnt+i+3]=IOPCommandData[i];
	}
	IOP_Cnt+=2;
	IOP_Cnt+=length;
	senddata[0]++;
}

//------------------ END OF File ---------------------
